/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <glib.h>
#include "nxplot-artft-font.h"
#include <fontconfig/fontconfig.h>
#include <libart_lgpl/art_affine.h>

#define GLYPH_CACHE_SIZE     32
#define ARTFTFONT_CACHE_SIZE 128
#define FONT_CACHE_SIZE      4

typedef struct {
    guint      refcount;
    GPtrArray *faces;		/* vector<FT_Face> */
    FcFontSet *all_fonts;
} Font;


struct _NxplotArtftFont {
    guint       refcount;
    Font       *font;
    double      transf[6];    /* General transforming matrix, including
			      * scaling. Format is that of libart,
			      * with translation being ignored. */
    GHashTable *glyph_cache; /* map<gunichar, NxplotArtftCachedGlyph> */
    gboolean    is_symbol;   /* TRUE for the "Symbol" font */
    FT_F26Dot6  char_width;
    FT_F26Dot6  char_height;
    FT_Matrix   ft_transf;   /* Transforming matrix, without scaling */
};

/* public variables */
FT_Render_Mode nxplot_artft_render_mode = ft_render_mode_normal;
guint          nxplot_artft_load_flags  = 0;


static GHashTable *fontcache = NULL; /* map<font_name_quark, Font> */
static FT_Library  ft_library;
static GHashTable *artftfont_cache = NULL; /* map<font_name_quark, GArray<NxplotArtftFont>> */
static GQuark      symbol_font_name;



static void               cleanup_font_cache             (void);
static void               cleanup_artftfont_cache        (void);
static void               nxplot_artft_cached_glyph_free (NxplotArtftCachedGlyph *glyph);
static void               cleanup_glyph_cache            (NxplotArtftFont        *font);
static FcFontSet*         fontmap_lookup                 (gchar const            *name);
static FT_Glyph           fontmap_lookup_char            (Font                   *font,
							  gunichar                chr,
							  FT_F26Dot6              char_width,
							  FT_F26Dot6              char_height);
static __inline__ FT_Face fontmap_lookup_try_face        (const char             *fname);



void nxplot_artft_font_init(void)
{
    int error;
    error = FT_Init_FreeType(&ft_library);
    if (error) g_error("Failed initialization of the FreeType library!");
    fontcache = g_hash_table_new(g_direct_hash, g_direct_equal);
    artftfont_cache = g_hash_table_new(g_direct_hash, g_direct_equal);
    symbol_font_name = g_quark_from_static_string("Symbol");
    FcInit();
}


static Font* get_font(GQuark name)
{
    Font *font;

    font = g_hash_table_lookup(fontcache, GINT_TO_POINTER(name));
    if (font) {
	++font->refcount;
	return font;
    } else {
	FT_Face face;
	int error, i;
	FcFontSet *fonts;
	FcChar8   *fname;

	fonts = fontmap_lookup(g_quark_to_string(name));
	g_return_val_if_fail(fonts, NULL);

	for (i = 0, face = NULL; i < fonts->nfont; ++i)
	{
	    fname = NULL;
	    FcPatternGetString(fonts->fonts[i],  FC_FILE, 0, &fname);
	    if (fname)
		face = fontmap_lookup_try_face(fname);
	    if (face) break;
	}

	g_return_val_if_fail(face, NULL);

	font = g_new0(Font, 1);
	font->refcount = 1;
	font->all_fonts = fonts;
	font->faces = g_ptr_array_new();
	g_ptr_array_add(font->faces, face);
	error = FT_Set_Pixel_Sizes(face, 0, 1);
#ifdef DEBUG
	if (error) g_warning("Failed to set pixel size of freetype font!");
#endif
	g_hash_table_insert(fontcache, GINT_TO_POINTER(name), font);

#if 0
	if (name == symbol_font_name) 
	{
	    FT_ULong  charcode;
	    FT_UInt   gindex;
	    char name[256];
	    FILE *list_file;
	    
	    g_message("Listing characters in the Symbol font...");
	    charcode = FT_Get_First_Char(face, &gindex);
	    list_file = fopen("/tmp/Symbol.txt", "w");
	    while (1)
	    {
		charcode = FT_Get_Next_Char(face, charcode, &gindex);
		if (gindex == 0) break;
		/* ... do something with (charcode,gindex) pair ... */
		FT_Get_Glyph_Name(face, gindex, name, 255);
		fprintf(list_file, "charcode=0x%08lx: glyph=%06i, name='%s'\n", charcode, gindex, name);
	    }
	    fclose(list_file);
	    g_message("Listing characters in the Symbol font... done.");
	}
#endif

	return font;
    }
}


static __inline__ void font_free(Font *font)
{
    int i;
    for (i = 0; i < font->faces->len; ++i)
	FT_Done_Face(g_ptr_array_index(font->faces, i));
    g_ptr_array_free(font->faces, TRUE);
    FcFontSetDestroy(font->all_fonts);
    g_free(font);
    /* g_message("Font @ %p has been freed", font); */
}


/* Get a glyph from the loaded faces in font; If the glyph is not
 * available, fontconfig is requested to search for a font containing
 * the required glyph, the font is loaded and the glyph is
 * obtained. */
static FT_Glyph font_get_glyph(Font *font, gunichar code,
			       FT_F26Dot6 char_width,
			       FT_F26Dot6 char_height)
{
    FT_Glyph glyph;
    FT_Face  face = NULL;
    int      i, error = 0;
    int      char_index;

    for (i = 0; i < font->faces->len; ++i)
    {
	face = (FT_Face) g_ptr_array_index(font->faces, i);	
	FT_Set_Char_Size(face, char_width, char_height, 0, 0);
	char_index = FT_Get_Char_Index(face, code);
	if (char_index == 0)
	    error = 1;
	else
	    error = FT_Load_Glyph(face, char_index, nxplot_artft_load_flags);
	if (!error) break;	/* found it! */
    }
    if (error) {
#ifdef DEBUG
	g_message("Trying to load alternate font for gunichar 0x%x"
		  " (already have %i faces)", code, font->faces->len);
#endif
	glyph = fontmap_lookup_char(font, code, char_width, char_height);
	if (glyph)
	    return glyph;	/* ... */
	else
	    error = 1;
    }
    if (error) {
	/* return "missing glyph" */
#ifdef DEBUG
	g_message("Load of alternate font for gunichar 0x%x failed!", code);
#endif
	face = g_ptr_array_index(font->faces, 0);
	FT_Set_Char_Size(face, char_width, char_height, 0, 0);
	FT_Load_Glyph(face, 0, nxplot_artft_load_flags);
	FT_Get_Glyph(face->glyph, &glyph);
	return glyph;
    } else {
	error = FT_Get_Glyph(face->glyph, &glyph);
	if (error) {
	    g_warning("FT_Get_Glyph");
	    return NULL;
	}
	return glyph;
    }
}


static void font_unref(Font *font)
{
    --font->refcount;
    if (font->refcount == 0) cleanup_font_cache();
}


static gboolean __inline__
_matrix_equal(double const *a, double const *b)
{
    return
	(a[0] == b[0]) &&
	(a[1] == b[1]) &&
	(a[2] == b[2]) &&
	(a[3] == b[3]);
}

void nxplot_artft_font_get_transf(NxplotArtftFont *font, FT_Matrix *transf)
{
    *transf = font->ft_transf;
}

/* Initialize font->char_width, font->char_height, font->ft_transf */
static void nxplot_artft_font_init_transf(NxplotArtftFont *font)
{
    double s;
    double scale_inv[6], rest[6];

    /*          T = R x S */
    /* T x INV(S) = R     */
    s = art_affine_expansion(font->transf);
    font->char_height = s*64;
    font->char_width  = font->char_height;
    s = font->char_height / 64.0;
    art_affine_scale(scale_inv, 1/s, 1/s);
    art_affine_multiply(rest, font->transf, scale_inv);
    
    /* (x1,  y1,  x2,  y2,  x3,  y3)  */
    /* (xx,  yx,  xy,  yy )           */
    /* ([0], [1], [2], [3], [4], [5]) */
    font->ft_transf.xx = rest[0]*0x10000;
    font->ft_transf.xy = rest[2]*0x10000;
    font->ft_transf.yx = rest[1]*0x10000;
    font->ft_transf.yy = rest[3]*0x10000;
}


NxplotArtftFont* nxplot_artft_get_font(GQuark name, double const *transf)
{
    GPtrArray       *array;
    NxplotArtftFont *artftfont, *artftfont1;
    int i;

    array = g_hash_table_lookup(artftfont_cache, GINT_TO_POINTER(name));
    if (!array) {
	array = g_ptr_array_new();
	g_hash_table_insert(artftfont_cache, GINT_TO_POINTER(name), array);
    }
    for (i = 0, artftfont = NULL; i < array->len; ++i) {
	artftfont1 = (NxplotArtftFont *) g_ptr_array_index(array, i);
	if (_matrix_equal(artftfont1->transf, transf)) {
	    artftfont = artftfont1;
	    break;
	}
    }
    if (artftfont) return artftfont;

    artftfont = g_new0(NxplotArtftFont, 1);
    g_ptr_array_add(array, artftfont);
    artftfont->refcount    = 1;
    artftfont->font        = get_font(name);
    memcpy(artftfont->transf, transf, 6*sizeof(double));
    nxplot_artft_font_init_transf(artftfont);
    artftfont->glyph_cache = g_hash_table_new(g_direct_hash, g_direct_equal);
    artftfont->is_symbol = (name == symbol_font_name);
    return artftfont;
}

static void nxplot_artft_font_free_helper(gpointer key,
					  gpointer value,
					  gpointer user_data)
{
    NxplotArtftCachedGlyph *glyph = (NxplotArtftCachedGlyph *) value;
    glyph->font = NULL;
    if (glyph->refcount != 0)
	g_warning("NxplotArtftCachedGlyph leak!");
    else
	nxplot_artft_cached_glyph_free(glyph);
}


static void nxplot_artft_font_free(NxplotArtftFont *font)
{
    g_hash_table_foreach(font->glyph_cache, nxplot_artft_font_free_helper, NULL);
    g_hash_table_destroy(font->glyph_cache);
    font_unref(font->font);
    g_free(font);
    /* g_message("NxplotArtftFont @ %p has been freed", font); */
}

void nxplot_artft_font_unref(NxplotArtftFont *font)
{
    --font->refcount;
    if (font->refcount == 0)
	cleanup_artftfont_cache();
}



NxplotArtftCachedGlyph* nxplot_artft_font_get_glyph(NxplotArtftFont *font, gunichar code)
{
    NxplotArtftCachedGlyph *glyph;

    if (font->is_symbol) {
	switch (code)
	{
	case 0x003BC: code = 0xB5; break;
	}
    }

    glyph = g_hash_table_lookup(font->glyph_cache, GINT_TO_POINTER(code));
    if (glyph)
	++glyph->refcount;
    else {
	glyph = g_new0(NxplotArtftCachedGlyph, 1);
	glyph->glyph = font_get_glyph(font->font, code,
				      font->char_width, font->char_height);
	if (glyph->glyph == NULL) {
	    g_free(glyph);
	    return NULL;
	}
	FT_Glyph_Transform(glyph->glyph, &font->ft_transf, NULL);
	/* glyph->index = glyph_index; */
	glyph->font  = font;
	glyph->refcount = 1;
	++font->refcount;
    }
    return glyph;
}


static void nxplot_artft_cached_glyph_free(NxplotArtftCachedGlyph *glyph)
{
    FT_Done_Glyph(glyph->glyph);
    nxplot_artft_font_unref(glyph->font);
    g_free(glyph);
/*     g_message("NxplotArtftCachedGlyph @ %p has been freed", glyph); */
	
}


void nxplot_artft_cached_glyph_unref(NxplotArtftCachedGlyph *glyph)
{
    --glyph->refcount;
    if (glyph->refcount == 0)
	cleanup_glyph_cache(glyph->font);
}


FT_Face nxplot_artft_font_get_face(NxplotArtftFont *font)
{
    return (FT_Face) g_ptr_array_index(font->font->faces, 0);
}



gchar** nxplot_artft_list_fonts(void)
{
    return NULL;
}

/* --- cache cleanup code --- */

static gboolean cleanup_font_cache_helper(gpointer key,
					  gpointer value,
					  gpointer user_data)
{
    Font  *font = (Font *) value;
    guint *size = (guint *) user_data;

    if (font->refcount == 0 && *size > FONT_CACHE_SIZE) {
	--*size;
	font_free(font);
	return TRUE;
    } else
	return FALSE;
}

/**
 * cleanup_font_cache:
 * 
 * Clears all Font's in the font cache the are unused until number of
 * items in cache does not exceed the cache size
 **/
static void cleanup_font_cache(void)
{
    guint size;
    if ((size = g_hash_table_size(fontcache)) <= FONT_CACHE_SIZE)
	return;
    g_hash_table_foreach_remove(fontcache, cleanup_font_cache_helper, &size);

}


static gboolean cleanup_artftfont_cache_helper(gpointer key,
					       gpointer value,
					       gpointer user_data)
{
    GPtrArray       *array = (GPtrArray *) value;
    NxplotArtftFont *font;
    guint len;
    int i;

    for (i = 0; len = array->len, i < len && len > ARTFTFONT_CACHE_SIZE ; ++i)
    {
	font = g_ptr_array_index(array, i);
	if (font->refcount == 0) {
	    g_ptr_array_remove_index_fast(array, i);
	    nxplot_artft_font_free(font);
	}
    }
    if (array->len == 0) {
	g_ptr_array_free(array, TRUE);
	return TRUE;
    } else
	return FALSE;
}

/**
 * cleanup_font_cache:
 * 
 * Clears all NxplotArtftFont's in the font cache the are unused until the number of
 * items in cache does not exceed the cache size
 **/
static void cleanup_artftfont_cache(void)
{
    guint size;
    if ((size = g_hash_table_size(artftfont_cache)) <= ARTFTFONT_CACHE_SIZE)
	return;
    g_hash_table_foreach_remove(artftfont_cache, cleanup_artftfont_cache_helper, &size);

}



static gboolean cleanup_glyph_cache_helper(gpointer key,
					   gpointer value,
					   gpointer user_data)
{
    NxplotArtftCachedGlyph *glyph = (NxplotArtftCachedGlyph *) value;
    guint *size = (guint *) user_data;

    if (glyph->refcount == 0 && *size > GLYPH_CACHE_SIZE) {
	--*size;
	nxplot_artft_cached_glyph_free(glyph);
	return TRUE;
    } else
	return FALSE;
}


static void cleanup_glyph_cache(NxplotArtftFont *font)
{
    guint size;
    if ((size = g_hash_table_size(font->glyph_cache)) <= GLYPH_CACHE_SIZE)
	return;
    g_hash_table_foreach_remove(font->glyph_cache, cleanup_glyph_cache_helper, &size);

}

/* Open font file and see if it supports unicode. Returns the FT_Face
 * if it meets the requirements, otherwise NULL. */
static __inline__ FT_Face fontmap_lookup_try_face(const char *fname)
{
    int     error;
    FT_Face face;
/*     g_message("Trying font %s", fname); */
    error = FT_New_Face(ft_library, fname, 0, &face);
    if (error == FT_Err_Unknown_File_Format) {
	g_warning("the font file %s could be opened and read, but it appears"
		  " that its font format is unsupported", fname);
	return NULL;
    } else if (error) {
	g_warning("the font file %s could not"
		  "be opened or read, or simply is broken", fname);
	return NULL;
    }
    error = FT_Select_Charmap(face, ft_encoding_unicode);
    if (error)  {
	g_warning("Failed to set unicode charmap of font %s!", fname);
#if 0
	const char *enc = NULL;
	int i;
	
	for (i = 0; i < font->face->num_charmaps; ++i)
	{
	    FT_CharMap charmap = font->face->charmaps[i];
		
	    switch (charmap->encoding)
	    {
#define _case(arg) case FT_ENCODING_##arg: enc = #arg; break
		_case(NONE);
		_case(UNICODE);
		_case(MS_SYMBOL);
		_case(ADOBE_LATIN_1);
		_case(ADOBE_STANDARD);
		_case(ADOBE_EXPERT);
		_case(ADOBE_CUSTOM);
		_case(APPLE_ROMAN);
		_case(OLD_LATIN_2);
		_case(MS_SJIS);
		_case(MS_GB2312);
		_case(MS_BIG5);
		_case(MS_WANSUNG);
		_case(MS_JOHAB);
#undef _case
	    default:
		enc = "unknown!";
	    }
	    g_message("Font has the charmap '%s'", enc);
	}
#endif
	FT_Done_Face(face);
	return NULL;
    }
    return face;
}

/* return true if the font appears to be from the ghostscript distribution */
static gboolean
_font_is_ghostript(FcPattern *font)
{
    FcChar8 *fname = NULL;

    FcPatternGetString(font,  FC_FILE, 0, &fname);
    return (fname && (strstr(fname, "ghostscript") || strstr(fname, "gsfonts")));
}

static FcFontSet* fontmap_lookup(gchar const  *name)
{
    gchar      *namedup = g_strdup(name);
    GString    *family  = g_string_new(NULL);
    int         weight  = FC_WEIGHT_MEDIUM;
    int         slant   = FC_SLANT_ROMAN;
    char       *token;
    FcPattern  *pattern, *symbfont;
    FcFontSet  *fonts, *newfonts;
    FcResult    result;
    int         i;

    token = strtok(namedup, " ");
    while (token)
    {
	if ((strcmp(token, "Regular") == 0))
	    slant = FC_SLANT_ROMAN;
	else if ((strcmp(token, "Roman") == 0))
	    slant = FC_SLANT_ROMAN;
	else if ((strcmp(token, "Italic") == 0))
	    slant = FC_SLANT_ITALIC;
	else if ((strcmp(token, "Oblique") == 0))
	    slant = FC_SLANT_OBLIQUE;
	else if ((strcmp(token, "Light") == 0))
	    weight = FC_WEIGHT_LIGHT;
	else if ((strcmp(token, "Medium") == 0))
	    weight = FC_WEIGHT_MEDIUM;
	else if ((strcmp(token, "Demi") == 0))
	    weight = FC_WEIGHT_DEMIBOLD;
	else if ((strcmp(token, "Bold") == 0))
	    weight = FC_WEIGHT_BOLD;
	else if ((strcmp(token, "Black") == 0))
	    weight = FC_WEIGHT_BLACK;
	else {
	    if (family->len)
		family = g_string_append_c(family, ' ');
	    family = g_string_append(family, token);
	}
	token = strtok(NULL, " ");
    }
    g_free(namedup);

    pattern = FcPatternCreate();
    FcPatternAddString(pattern, FC_FAMILY, family->str);
    FcPatternAddInteger(pattern, FC_SLANT, slant);
    FcPatternAddInteger(pattern, FC_WEIGHT, weight);
    FcConfigSubstitute(NULL, pattern, FcMatchPattern);
    fonts = FcFontSort(NULL, pattern, FcFalse, NULL, &result);
    FcPatternDestroy(pattern);

    if (strcmp("Standard Symbols L", family->str) == 0)
	/* if requested font is already symbol, don't insert another copy */
	symbfont = NULL;
    else {
        FcFontSet *fonts;
	/* lookup the symbols font */
	pattern = FcPatternCreate();
	FcPatternAddString(pattern, FC_FAMILY, "Standard Symbols L");
	FcPatternAddInteger(pattern, FC_SLANT, FC_SLANT_ROMAN);
	FcPatternAddInteger(pattern, FC_WEIGHT, FC_WEIGHT_MEDIUM);
	FcDefaultSubstitute(pattern);
        fonts = FcFontSort(NULL, pattern, FcFalse, NULL, &result);
        symbfont = NULL;
        for (i = 0; i < fonts->nfont; ++i) {
            FcChar8 *fname = NULL;
            if (_font_is_ghostript(fonts->fonts[i])) {
                symbfont = fonts->fonts[i];
                FcPatternReference(symbfont);
                break;
            }
            FcPatternGetString(fonts->fonts[i],  FC_FILE, 0, &fname);
            g_message("Symbol font '%s' doesn't look from ghostscript", fname);
        }
	if (!symbfont)
	    g_warning("The font 'Standard Symbols L' distributed by ghostscript"
                      " could not be found."
		      "  Please ask your admin to correct the fontconfig"
		      " configuration in order to include this font, "
		      "which usually can found among the ghostscript fonts");
	FcPatternDestroy(pattern);
        FcFontSetDestroy(fonts);
    }

    g_string_free(family, TRUE);


    {    /* reorder fonts, putting ghostscript fonts in first place */
        FcFontSet *tmpfonts;
        
        tmpfonts = FcFontSetCreate();
        for (i = 0; i < fonts->nfont; ++i) {
            if (_font_is_ghostript(fonts->fonts[i])) {
                FcPatternReference(fonts->fonts[i]);
                FcFontSetAdd(tmpfonts, fonts->fonts[i]);
            }
        }
        for (i = 0; i < fonts->nfont; ++i) {
            if (!(_font_is_ghostript(fonts->fonts[i]))) {
                FcPatternReference(fonts->fonts[i]);
                FcFontSetAdd(tmpfonts, fonts->fonts[i]);
            }
        }
        FcFontSetDestroy(fonts);
        fonts = tmpfonts;
    }

    /* create a new font list, inserting the symbols font in second position */
    newfonts = FcFontSetCreate();
    for (i = 0; ; ++i) {
	if (i == 1 && symbfont)
	    FcFontSetAdd(newfonts, symbfont);
	if (i >= fonts->nfont)
	    break;
	FcPatternReference(fonts->fonts[i]);
	FcFontSetAdd(newfonts, fonts->fonts[i]);
    }
    FcFontSetDestroy(fonts);

    return newfonts;
}


/* Try to load any additional font that contains the given unicode character */
static FT_Glyph fontmap_lookup_char(Font *font,
				    gunichar chr,
				    FT_F26Dot6 char_width,
				    FT_F26Dot6 char_height)
{
    FT_Glyph glyph = NULL;
    int      char_index, error, i;
    FT_Face  face;

    for (i = 0; i < font->all_fonts->nfont; ++i)
    {
	FcCharSet *cs     = NULL;
	FcChar8   *fname  = NULL;
	FcPattern *fcfont = font->all_fonts->fonts[i];

	FcPatternGetCharSet(fcfont, FC_CHARSET, 0, &cs);
	if (cs == NULL) {
#ifdef DEBUG
	    FcChar8 *name = NULL;
	    FcPatternGetString(fcfont,  FC_FAMILY, 0, &name);
	    g_warning("Font %s has no charset", name);
#endif
	    continue;
	}
	if (!FcCharSetHasChar(cs, chr)) {
#ifdef DEBUG
	    FcChar8 *name = NULL;
	    FcPatternGetString(fcfont,  FC_FAMILY, 0, &name);
	    g_warning("Font %s doesn't have char 0x%x", name, chr);
#endif
	    continue;
	}

	FcPatternGetString(fcfont,  FC_FILE, 0, &fname);
	if (!fname) continue;
	face = fontmap_lookup_try_face(fname);
	if (face) {
	    FT_Set_Char_Size(face, char_width, char_height, 0, 0);
	    char_index = FT_Get_Char_Index(face, chr);
	    if (char_index == 0)
		error = 1;
	    else
		error = FT_Load_Glyph(face, char_index, nxplot_artft_load_flags);
	    if (error) {
		g_warning("fontconfig told us that font %s had char 0x%x,"
			  " but apparently that isn't true", fname, chr);
		FT_Done_Face(face);
	    }
	    FT_Get_Glyph(face->glyph, &glyph);
	    g_ptr_array_add(font->faces, face);
	    return glyph;
	} else {
#ifdef DEBUG
	    FcChar8 *name = NULL;
	    FcPatternGetString(fcfont,  FC_FAMILY, 0, &name);
	    g_warning("Font %s (%s) could not be opened", name, fname);
#endif
	    continue;
	}
    }
    return NULL;
}

