/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "nxplot-artft.h"
#include <math.h>
#include <string.h>
#include <libart_lgpl/art_rgba.h>

#ifdef HAVE_HYPOT
# define myhypot(x, y) hypot((x), (y))
#else
# define myhypot(x, y) sqrt((x)*(x) + (y)*(y))
#endif

#define NXPLOT_ARTFT_ENABLE_KERNING


/* signals */
enum {
  UPDATE,
  LAST_SIGNAL
};


static guint signals[LAST_SIGNAL] = { 0 };
static NxplotArtftClass *nxplot_artft_class = NULL;
static GQuark            symbol_font_name;


static void               nxplot_artft_init                 (NxplotArtft      *instance,
							     gpointer          g_class);
static void               nxplot_artft_class_base_init      (NxplotArtftClass *class);
static void               nxplot_artft_class_init           (NxplotArtftClass *class,
							     gpointer          data);
static void               nxplot_artft_shutdown             (GObject          *object);
static void               nxplot_artft_update_font_size     (NxplotArtft      *self);
static NxplotArtftState*  nxplot_artft_state_new            (void);
static void               nxplot_artft_state_free           (NxplotArtftState *state);
static void               nxplot_artft_state_copy           (NxplotArtftState *dest,
							     NxplotArtftState *src);
static NxplotArtftState * nxplot_artft_state_dup            (NxplotArtftState *state);
static void               nxplot_artft_clear_states         (Nxplot           *nxplot);
static void               nxplot_artft_invalidate_font      (NxplotArtft      *self);
static void               nxplot_artft_ensure_font          (NxplotArtft      *self);
static inline void        nxplot_artft_compute_space2window (NxplotArtft      *self,
							     double const     *space2unit,
							     double           *space2window);
static void               update_clipping                   (NxplotArtft      *self);
static void               widget_resized                    (GtkWidget        *widget,
							     GtkAllocation    *allocation,
							     NxplotArtft      *self);
static void               widget_realized                   (GtkWidget        *widget,
							     NxplotArtft      *self);

/* --- driver implementations --- */
static void   nxplot_artft_open          (Nxplot               *nxplot);
static void   nxplot_artft_close         (Nxplot               *nxplot);
static void   nxplot_artft_endpath       (Nxplot               *nxplot);
static void   nxplot_artft_erase         (Nxplot               *nxplot);
static void   nxplot_artft_line          (Nxplot               *nxplot,
					  double                x1,
					  double                y1,
					  double                x2,
					  double                y2);
static void   nxplot_artft_space         (Nxplot               *nxplot,
					  double                x1,
					  double                y1,
					  double                x2,
					  double                y2);
static void   nxplot_artft_linerel       (Nxplot               *nxplot,
					  double                x1,
					  double                y1,
					  double                x2,
					  double                y2);
static void   nxplot_artft_linewidth     (Nxplot               *nxplot,
					  double                width);
static void   nxplot_artft_pencolor      (Nxplot               *nxplot,
					  guint16               red,
					  guint16               green,
					  guint16               blue);
static void   nxplot_artft_bgcolor       (Nxplot               *nxplot,
					  guint16               red,
					  guint16               green,
					  guint16               blue);
static void   nxplot_artft_pencolorname  (Nxplot               *nxplot,
					  const char           *name);
static void   nxplot_artft_move          (Nxplot               *nxplot,
					  double                x,
					  double                y);
static void   nxplot_artft_moverel       (Nxplot               *nxplot,
					  double                x,
					  double                y);
static void   nxplot_artft_fontname      (Nxplot               *nxplot,
					  const char           *font_name);
static void   _list_fonts                (Nxplot               *nxplot,
					  gchar              ***fontnames);
static void   nxplot_artft_fontsize      (Nxplot               *nxplot,
					  double                size);
static void   nxplot_artft_textangle     (Nxplot               *nxplot,
					  double                angle);
static void   nxplot_artft_linemode      (Nxplot               *nxplot,
					  NxplotLinemode        mode);
static void   nxplot_artft_label         (Nxplot               *nxplot,
					  NxplotHAlign          halign,
					  NxplotVAlign          valign,
					  const char           *text);
static double nxplot_artft_labelwidth    (Nxplot               *nxplot,
					  const char           *text);
static double nxplot_artft_descender     (Nxplot               *nxplot);
static void   nxplot_artft_circle        (Nxplot               *nxplot,
					  double                xc,
					  double                yc,
					  double                r);
static void   nxplot_artft_savestate     (Nxplot               *nxplot);
static void   nxplot_artft_restorestate  (Nxplot               *nxplot);
static void   nxplot_artft_setmatrix     (Nxplot               *nxplot,
					  double                m0,
					  double                m1,
					  double                m2,
					  double                m3,
					  double                tx,
					  double                ty);
static void   nxplot_artft_concat        (Nxplot               *nxplot,
					  double                m0,
					  double                m1,
					  double                m2,
					  double                m3,
					  double                tx,
					  double                ty);
static void   nxplot_artft_rotate        (Nxplot               *nxplot,
					  double                theta);
static void   nxplot_artft_scale         (Nxplot               *nxplot,
					  double                sx,
					  double                sy);
static void   nxplot_artft_translate     (Nxplot               *nxplot,
					  double                tx,
					  double                ty);
static void   nxplot_artft_space2window  (Nxplot               *nxplot,
					  double                x,
					  double                y,
					  double               *wx,
					  double               *wy);
static void   nxplot_artft_window2space  (Nxplot               *nxplot,
					  double                wx,
					  double                wy,
					  double               *x,
					  double               *y);
static void   nxplot_artft_clear_rect    (Nxplot               *nxplot,
					  double                x1,
					  double                y1,
					  double                x2,
					  double                y2);
static void   nxplot_artft_setcliprect   (Nxplot               *nxplot,
					  NxplotRect           *clip_rect);
static void   nxplot_artft_labelmetrics  (Nxplot               *nxplot,
					  const char           *text,
					  NxplotLabelMetrics   *metrics);
static void   nxplot_artft_fillcolor     (Nxplot               *nxplot,
					  guint16               red,
					  guint16               green,
					  guint16               blue);
static void   nxplot_artft_fillcolorname (Nxplot               *nxplot,
					  const char           *name);
static void   nxplot_artft_fillopacity   (Nxplot               *nxplot,
					  guint16               opacity);



static GTypeInfo typeinfo = 
{
    /* interface types, classed types, instantiated types */
    sizeof(NxplotArtftClass),

    (GBaseInitFunc)nxplot_artft_class_base_init,
    NULL, // GBaseFinalizeFunc      base_finalize;
    
    /* classed types, instantiated types */
    (GClassInitFunc)nxplot_artft_class_init,
    NULL, // GClassFinalizeFunc     class_finalize;
    NULL, // gconstpointer          class_data;
    
    /* instantiated types */
    sizeof(NxplotArtft),		//guint16                instance_size;
    0,					//guint16                n_preallocs;
    (GInstanceInitFunc)nxplot_artft_init, //GInstanceInitFunc      instance_init;
    
    /* value handling */
    NULL,				//const GTypeValueTable	*value_table;
};



GType nxplot_artft_get_type(void)
{
    static GType type = 0;
    if (!type)
    {
	type = g_type_register_static(nxplot_get_type(), "NxplotArtft", 
				      &typeinfo, 0);
    }
    return type;
}


static void nxplot_artft_init(NxplotArtft *self, gpointer g_class)
{
    if (!nxplot_artft_class)
	nxplot_artft_class = NXPLOT_ARTFT_GET_CLASS(self);
    self->saved_states = g_queue_new();
    self->state.linewidth = 0.01;
    self->state.foreground  = 0;
    self->state.background  = 0xffffffff; /* white background by default */
    self->state.fillcolor   = 0xffffffff;
    self->state.fillopacity = 0; /* unfilled by default */
    self->state.linemode = NXPLOT_LINEMODE_SOLID;
    self->state.joinmode = ART_PATH_STROKE_JOIN_BEVEL;
    self->state.capmode = ART_PATH_STROKE_CAP_BUTT;
    self->vpath_array = g_array_new(FALSE, FALSE, sizeof(ArtVpath));
    self->state.textangle = 0;
    self->state.fontsize = 0.05;
    memset(&self->allocation, 0, sizeof(self->allocation));
    self->clipbuf = g_new0(ArtPixBuf, 1);

    nxplot_artft_space(NXPLOT(self), 0, 0, 1, 1);
    
    self->using_symbol_font = FALSE;
    self->font = NULL;
    nxplot_artft_fontname(NXPLOT(self), "Sans Regular");
    nxplot_artft_update_font_size(self);
    nxplot_artft_invalidate_font(self);
}


static void nxplot_artft_class_base_init(NxplotArtftClass *class)
{
}


static void nxplot_artft_class_init(NxplotArtftClass *class, gpointer data)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    NxplotClass *nxplot_class = NXPLOT_CLASS(class);

    object_class->finalize	= nxplot_artft_shutdown;
    nxplot_class->open		= nxplot_artft_open;
    nxplot_class->close		= nxplot_artft_close;
    nxplot_class->endpath	= nxplot_artft_endpath;
    nxplot_class->erase		= nxplot_artft_erase;
    nxplot_class->clearrect	= nxplot_artft_clear_rect;
    nxplot_class->line		= nxplot_artft_line;
    nxplot_class->space		= nxplot_artft_space;
    nxplot_class->linerel	= nxplot_artft_linerel;
    nxplot_class->circle	= nxplot_artft_circle;
    nxplot_class->linewidth	= nxplot_artft_linewidth;
    nxplot_class->pencolor	= nxplot_artft_pencolor;
    nxplot_class->bgcolor	= nxplot_artft_bgcolor;
    nxplot_class->pencolorname	= nxplot_artft_pencolorname;
    nxplot_class->move		= nxplot_artft_move;
    nxplot_class->moverel	= nxplot_artft_moverel;
    nxplot_class->linemode	= nxplot_artft_linemode;
    nxplot_class->textangle	= nxplot_artft_textangle;
    nxplot_class->fontname	= nxplot_artft_fontname;
    nxplot_class->listfonts	= _list_fonts;
    nxplot_class->fontsize	= nxplot_artft_fontsize;
    nxplot_class->label	        = nxplot_artft_label;
    nxplot_class->labelwidth    = nxplot_artft_labelwidth;
    nxplot_class->descender     = nxplot_artft_descender;
    nxplot_class->labelmetrics  = nxplot_artft_labelmetrics;

    nxplot_class->savestate	= nxplot_artft_savestate;
    nxplot_class->restorestate	= nxplot_artft_restorestate;
    nxplot_class->setmatrix	= nxplot_artft_setmatrix;
    nxplot_class->concat	= nxplot_artft_concat;
    nxplot_class->rotate	= nxplot_artft_rotate;
    nxplot_class->scale		= nxplot_artft_scale;
    nxplot_class->translate	= nxplot_artft_translate;

    nxplot_class->space2window	= nxplot_artft_space2window;
    nxplot_class->window2space	= nxplot_artft_window2space;
    nxplot_class->setcliprect	= nxplot_artft_setcliprect;

    nxplot_class->fillcolor	= nxplot_artft_fillcolor;
    nxplot_class->fillcolorname	= nxplot_artft_fillcolorname;
    nxplot_class->fillopacity	= nxplot_artft_fillopacity;

    signals[UPDATE] = class->update_signal = 
	g_signal_newv("update",
		      NXPLOT_ARTFT_CLASS_TYPE(class),
		      G_SIGNAL_RUN_LAST, NULL, NULL, NULL,
		      g_cclosure_marshal_VOID__VOID,
		      G_TYPE_NONE, 0, NULL);

    nxplot_artft_font_init();
    symbol_font_name = g_quark_from_static_string("Standard Symbols L Regular");
}


static void nxplot_artft_cleanup(NxplotArtft *self)
{
    if (self->pixmap) {
	g_object_unref(self->pixmap);
	self->pixmap = NULL;
    }
    if (self->pixbuf) {
	art_pixbuf_free(self->pixbuf);
	self->pixbuf = NULL;
    }
}


static void widget_destroyed(NxplotArtft *self)
{
    g_assert(IS_NXPLOT_ARTFT(self));
    nxplot_artft_cleanup(self);
    self->widget = NULL;
}


static gint widget_expose_event(GtkWidget *widget, GdkEventExpose *event, NxplotArtft *self)
{
    gint offset_x, offset_y;
    g_assert(IS_NXPLOT_ARTFT(self));
    if (GTK_WIDGET_NO_WINDOW(widget)) {
	offset_x = self->allocation.x;
	offset_y = self->allocation.y;
    } else {
	offset_x = 0;
	offset_y = 0;
    }
#if 1
    gdk_draw_rgb_image(widget->window, widget->style->black_gc, 
		       event->area.x, event->area.y, event->area.width, 
		       event->area.height, GDK_RGB_DITHER_NONE, 
		       self->pixbuf->pixels + (event->area.y - offset_y) *
		       self->pixbuf->rowstride + 3*(event->area.x - offset_x),
		       self->pixbuf->rowstride);
#else
    /* FIXME: this is no longer working because of offset_[xy] */
    if (self->pixmap && widget->window)
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
			self->pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
#endif
    return TRUE;
}


static void nxplot_artft_shutdown(GObject *object)
{
    NxplotArtft *self = NXPLOT_ARTFT(object);
    gulong sig;
    /* g_message(__FUNCTION__); */
    nxplot_artft_cleanup(self);
    if (self->widget) {
	gtk_object_weakunref(GTK_OBJECT(self->widget),
			     (GtkDestroyNotify)widget_destroyed, self);
	/* disconnect expose signal */
	sig = g_signal_handler_find(self->widget,
				    G_SIGNAL_MATCH_FUNC|G_SIGNAL_MATCH_DATA,
				    0, 0, NULL, widget_expose_event, object);
	if (sig) g_signal_handler_disconnect(self->widget, sig);
	/* disconnect size-allocate signal */
	sig = g_signal_handler_find(self->widget,
				    G_SIGNAL_MATCH_FUNC|G_SIGNAL_MATCH_DATA,
				    0, 0, NULL, widget_resized, object);
	if (sig) g_signal_handler_disconnect(self->widget, sig);
	/* disconnect realize signal */
	sig = g_signal_handler_find(self->widget,
				    G_SIGNAL_MATCH_FUNC|G_SIGNAL_MATCH_DATA,
				    0, 0, NULL, widget_realized, object);
	if (sig) g_signal_handler_disconnect(self->widget, sig);
    }
    /* free list of saved states */
    while (!g_queue_is_empty(self->saved_states))
    {
	NxplotArtftState *state;
	state = g_queue_pop_tail(self->saved_states);
	nxplot_artft_state_free(state);
    }
    g_queue_free(self->saved_states);
    self->saved_states = NULL;
    self->widget = NULL;
    g_array_free(self->vpath_array, TRUE);
#if 0
    if (self->dirty_rects) {
	gdk_region_destroy(self->dirty_rects);
	self->dirty_rects = NULL;
    }
#endif
    if (self->clipbuf) {
	g_free(self->clipbuf);
	self->clipbuf = NULL;
    }
}


static struct {
    int ndashes;
    double dashes[8];
} line_modes[7] = 
{
    {0, {0,0,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_SOLID */
    {2, {1,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_DOTTED */
    {4, {1,2,4,2,0,0,0,0}},	/* NXPLOT_LINEMODE_DOTDASHED */
    {2, {3,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_SHORTDASHED */
    {2, {6,2,0,0,0,0,0,0}},	/* NXPLOT_LINEMODE_LONGDASHED */
    {6, {1,2,1,2,4,2,0,0}},	/* NXPLOT_LINEMODE_DOTDOTDASHED */
    {8, {1,2,1,2,1,2,4,2}}	/* NXPLOT_LINEMODE_DOTDOTDOTDASHED */
};

static inline void update_dashes(NxplotArtft *self)
{
    double scale = self->state.linewidth;
    int i;
    if (self->state.linemode > 0 && self->state.linemode < 7) {
	/* copy the template dashes to self while scaling the dashes
           by linewidth */
	self->dash_style.offset = 0;
	self->dash_style.n_dash = line_modes[self->state.linemode].ndashes;
	self->dash_style.dash = self->dashes;
	for (i = 0; i < 8; i++)
	    self->dashes[i] = line_modes[self->state.linemode].dashes[i] * scale;
    }
}

static inline void update_wlinewidth(NxplotArtft *self)
{
    double dx, dy;
    dx = NXPLOT(self)->view.x2 - NXPLOT(self)->view.x1;
    dy = NXPLOT(self)->view.y2 - NXPLOT(self)->view.y1;
    self->wlinewidth = MIN(self->allocation.width/dx, self->allocation.height/dy) *
	self->state.linewidth;
}

#if 0
static void add_dirty_rect(NxplotArtft *self, gint x, gint y, gint width, gint height)
{
    GdkRectangle rect;
    rect.x = x; rect.y = y;
    rect.width = width; rect.height = height;
    if (!self->dirty_rects)
	self->dirty_rects = gdk_region_rectangle(&rect);
    else
	gdk_region_union_with_rect(self->dirty_rects, &rect);
}


static void nxplot_artft_flush(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    
}
#endif

static inline void nxplot_artft_compute_space2window(NxplotArtft *self,
						     double const *space2unit,
						     double *space2window)
{
    double tmp[6];

    art_affine_scale(tmp, 1.0, -1.0);
    art_affine_multiply(self->space2window, self->state.space2unit, tmp);
    /* this is weird; I don't remember why I did this translation... :| */
    art_affine_translate(tmp, 0.0, 1.0);
    art_affine_multiply(space2window, space2window, tmp);
    art_affine_scale(tmp, self->allocation.width, self->allocation.height);
    art_affine_multiply(space2window, space2window, tmp);
}

static inline void nxplot_artft_update_affine(NxplotArtft *self)
{
    /* recalculate the space-to-window affine transformation */
    nxplot_artft_compute_space2window(self, self->state.space2unit,
				      self->space2window);
    /* calculate window2space matrix for reverse mapping */
    art_affine_invert(self->window2space, self->space2window);
}

static void size_update(NxplotArtft *self)
{
    if (self->widget) {
	if (!GTK_WIDGET_REALIZED(self->widget)) {
	    g_signal_connect_after(self->widget, "realize",
				   G_CALLBACK(widget_realized), self);
	    return;
	}
    }
    if (NXPLOT(self)->opened) nxplot_close(NXPLOT(self));
    nxplot_artft_cleanup(self);

    self->pixbuf = art_pixbuf_new_rgb
	(art_new(art_u8, 3*self->allocation.width*self->allocation.height),
	 self->allocation.width, self->allocation.height,
	 3*self->allocation.width);
/*      self->pixmap = gdk_pixmap_new(GTK_WIDGET(self->widget)->window,  */
/*  				  self->width, self->height, -1); */
    nxplot_artft_update_affine(self);
    update_wlinewidth(self);
    nxplot_artft_update_font_size(self);
    update_clipping(self);
//    update_dashes(self);

    g_signal_emit(self, NXPLOT_ARTFT_GET_CLASS(self)->update_signal, 0);
}


static void widget_realized(GtkWidget *widget, NxplotArtft *self)
{
    g_assert(IS_NXPLOT_ARTFT(self));
    size_update(self);
}


static void widget_resized(GtkWidget *widget, GtkAllocation *allocation,
			   NxplotArtft *self)
{
    g_return_if_fail(IS_NXPLOT_ARTFT(self));
    self->allocation  = *allocation;
/*     g_message("nxplot-artft: widget allocation is (%i, %i, %i, %i)", */
/* 	      allocation->x, allocation->y, allocation->width, allocation->height); */
    if (GTK_WIDGET_REALIZED(self->widget))
	size_update(self);
    else
	g_signal_connect_after(self->widget, "realize", G_CALLBACK(widget_realized), self);
}


void nxplot_artft_set_size(NxplotArtft *self, int width, int height)
{
    g_return_if_fail(IS_NXPLOT_ARTFT(self));
    g_return_if_fail(self->widget == NULL);
    g_return_if_fail(width > 0);
    g_return_if_fail(height > 0);

    self->allocation.x = 0;
    self->allocation.y = 0;
    self->allocation.width  = width;
    self->allocation.height = height;
    size_update(self);
}


Nxplot* nxplot_artft_new_widgetless(int width, int height)
{
    GTypeInstance *instance;
    NxplotArtft *self;
    Nxplot *nxplot;

    instance = g_type_create_instance(nxplot_artft_get_type());
    self = NXPLOT_ARTFT(instance);
    nxplot = NXPLOT(instance);
    self->widget = NULL;
    nxplot_artft_set_size(self, width, height);

    return NXPLOT(self);
}


Nxplot* nxplot_artft_new(GtkWidget *widget)
{
    GTypeInstance *instance;
    NxplotArtft *self;
    Nxplot *nxplot;

    g_return_val_if_fail(GTK_IS_WIDGET(widget), NULL);
    instance = g_type_create_instance(nxplot_artft_get_type());
    self = NXPLOT_ARTFT(instance);
    nxplot = NXPLOT(instance);

    self->widget = GTK_WIDGET(widget);
    gtk_widget_set_double_buffered(self->widget, FALSE);

    gtk_object_weakref(GTK_OBJECT(widget), (GtkDestroyNotify)widget_destroyed, self);
    gtk_signal_connect(GTK_OBJECT(widget), "size-allocate",
		       GTK_SIGNAL_FUNC(widget_resized), self);
    gtk_signal_connect(GTK_OBJECT(widget), "expose-event",
		       GTK_SIGNAL_FUNC(widget_expose_event), self);
    gtk_widget_queue_resize(GTK_WIDGET(widget));
    return NXPLOT(self);
}


static void nxplot_artft_open(Nxplot *nxplot)
{
/*      NxplotArtft *self = NXPLOT_ARTFT(nxplot); */
    if(nxplot->opened) return;
    /* (...) */
    nxplot->opened = TRUE;
}


/* Render the rgb buffer into the backing-store pixmap. The latter
   will be rendered into the widget during expose events. */
G_GNUC_UNUSED static void nxplot_artft_render_rgb(NxplotArtft *self) 
{
    g_return_if_fail(self->pixbuf);
    g_return_if_fail(self->pixmap);
    gdk_draw_rgb_image(self->pixmap, self->widget->style->black_gc, 
		       0, 0, self->pixbuf->width, 
		       self->pixbuf->height, GDK_RGB_DITHER_NONE,
		       self->pixbuf->pixels, self->pixbuf->rowstride);
    gtk_widget_queue_draw(self->widget);
}


static void nxplot_artft_free_points(NxplotArtft *self)
{
    self->vpath_array = g_array_set_size(self->vpath_array, 0);
    self->inpath = FALSE;
}


static inline void nxplot_artft_add_point(NxplotArtft *self,
					  ArtPathcode code, double x, double y)
{
    ArtVpath vpath;
    vpath.code = code;
    vpath.x = x;
    vpath.y = y;
    self->vpath_array = g_array_append_val(self->vpath_array, vpath);
    self->inpath = TRUE;
}


static void nxplot_artft_endpath(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->clipbuf);
    if (self->vpath_array->len)
    {
	ArtSVP *svp;
	ArtVpath *vpath, *vpath1;

	nxplot_artft_add_point(self, ART_END, self->state.last_x, self->state.last_y);
	if (self->state.fillopacity > 0)
	{
	    guint32 rgba = (self->state.fillcolor & 0xffffff00) |
		(self->state.fillopacity >> 8);
	    vpath = art_vpath_affine_transform((ArtVpath*)self->vpath_array->data,
					       self->space2window);
	    svp = art_svp_from_vpath(vpath);
	    art_free(vpath);
	    art_rgb_svp_alpha(svp, self->clip_x, self->clip_y,
			      self->clip_x + self->clip_w,
			      self->clip_y + self->clip_h,
			      rgba, self->clipbuf->pixels,
			      self->clipbuf->rowstride, NULL);
	    art_free(svp);
	}

	if (self->state.linemode != NXPLOT_LINEMODE_SOLID && 
	    self->state.linemode != NXPLOT_LINEMODE_DISCONNECTED) {
	    vpath1 = art_vpath_dash((ArtVpath*)self->vpath_array->data, &self->dash_style);
	    vpath = art_vpath_affine_transform(vpath1, self->space2window);
	    art_free(vpath1);
	} else 
	    vpath = art_vpath_affine_transform((ArtVpath*)self->vpath_array->data,
					       self->space2window);
	svp = art_svp_vpath_stroke(vpath, self->state.joinmode, self->state.capmode,
				   self->wlinewidth, 1.0, .5);

	/* Since libplot and postscript don't support alpha, we don't
           support it either, but it would be easy. */
	art_rgb_svp_alpha(svp, self->clip_x, self->clip_y,
			  self->clip_x + self->clip_w,
			  self->clip_y + self->clip_h,
			  self->state.foreground|0xff, self->clipbuf->pixels,
			  self->clipbuf->rowstride, NULL);
	art_free(svp);
	art_free(vpath);
    }
    nxplot_artft_free_points(self);
}


static void nxplot_artft_close(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    if (!nxplot->opened) return;
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    nxplot_artft_clear_states(nxplot);
    nxplot_artft_space(nxplot, nxplot->view.x1, nxplot->view.y1,
		       nxplot->view.x2, nxplot->view.y2);
//    nxplot_artft_render_rgb(self);
    nxplot->opened = FALSE;
}


static void nxplot_artft_erase(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);

    if (!self->pixbuf) return;

    g_assert(self->pixbuf->height > 0);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);

    art_rgb_fill_run(self->pixbuf->pixels,
		     (self->state.background >> 24) & 0xff,
		     (self->state.background >> 16) & 0xff,
		     (self->state.background >>  8) & 0xff,
		     self->pixbuf->width*self->pixbuf->height);

    if (self->widget)
	gtk_widget_queue_draw(self->widget);
}


static void nxplot_artft_line(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath) {
	if (self->state.last_x == x1 && self->state.last_y == y1)
	    /* line does not interrupt path */
	    nxplot_artft_add_point(self, ART_LINETO, x2, y2);
	else {
	    nxplot_artft_endpath(NXPLOT(self));
	    nxplot_artft_add_point(self, ART_MOVETO, x1, y1);
	    nxplot_artft_add_point(self, ART_LINETO, x2, y2);
	}
    } else {
	nxplot_artft_add_point(self, ART_MOVETO, x1, y1);
	nxplot_artft_add_point(self, ART_LINETO, x2, y2);
    }
    self->state.last_x = x2;
    self->state.last_y = y2;
}


static void nxplot_artft_space(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double tmp[6];

    if (self->inpath)
	nxplot_artft_endpath(nxplot);

    nxplot->view.x1 = x1;
    nxplot->view.y1 = y1;
    nxplot->view.x2 = x2;
    nxplot->view.y2 = y2;

    art_affine_translate(self->state.space2unit, -x1, -y1);
    art_affine_scale(tmp, 1/(x2 - x1), 1/(y2 - y1));
    art_affine_multiply(self->state.space2unit, self->state.space2unit, tmp);
    nxplot_artft_update_affine(self);
    update_wlinewidth(self);
    nxplot_artft_update_font_size(self);
}


static void nxplot_artft_linerel(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    nxplot_artft_line(nxplot, self->state.last_x + x1, self->state.last_y + y1,
		      self->state.last_x + x2, self->state.last_y + y2);
}


static void nxplot_artft_linewidth(Nxplot *nxplot, double width)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    g_return_if_fail(width > 0);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    self->state.linewidth = width;
    update_wlinewidth(self);
    update_dashes(self);
}


static void nxplot_artft_pencolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    guint32 r, g, b;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    r = red; g = green; b = blue;
    self->state.foreground = ((r<<16)&0xff000000) | ((g<<8)&0x00ff0000) | ((b)&0x0000ff00);
}


static void nxplot_artft_bgcolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    guint32 r, g, b;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    r = red; g = green; b = blue;
    self->state.background = ((r<<16)&0xff000000) | ((g<<8)&0x00ff0000) | ((b)&0x0000ff00);
}


static void nxplot_artft_pencolorname(Nxplot *nxplot, const char *name)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    GdkColor color = {0,0,0,0};
    gint rv;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    
    rv = gdk_color_parse(name, &color);
    if (!rv) g_warning("Parsing of color '%s' failed", name);
    nxplot_artft_pencolor(nxplot, color.red, color.green, color.blue);
}


static void nxplot_artft_move(Nxplot *nxplot, double x, double y)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath) nxplot_artft_endpath(nxplot);
    self->state.last_x = x;
    self->state.last_y = y;
}


static void nxplot_artft_circle(Nxplot *nxplot, double xc, double yc, double r)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    ArtVpath *vpath, *vpath1;
    guint n;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    vpath = art_vpath_new_circle(xc, yc, r);
    for (vpath1 = vpath, n = 0; vpath1->code != ART_END; ++n, ++vpath1);
    self->vpath_array = g_array_append_vals(self->vpath_array,
					    vpath, n);
    art_free(vpath);
    nxplot_artft_endpath(nxplot);
    self->state.last_x = xc;
    self->state.last_y = yc;
}


static void nxplot_artft_moverel(Nxplot *nxplot, double x, double y)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    nxplot_artft_move(nxplot, self->state.last_x + x, self->state.last_y + y);
}


static void nxplot_artft_linemode(Nxplot *nxplot, NxplotLinemode mode)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    self->state.linemode = mode;
    switch (mode)
    {
    case NXPLOT_LINEMODE_SOLID:
	break;
    case NXPLOT_LINEMODE_DOTTED:
    case NXPLOT_LINEMODE_DOTDASHED:
    case NXPLOT_LINEMODE_SHORTDASHED:
    case NXPLOT_LINEMODE_LONGDASHED:
    case NXPLOT_LINEMODE_DOTDOTDASHED:
    case NXPLOT_LINEMODE_DOTDOTDOTDASHED:
	update_dashes(self);
	break;
    case NXPLOT_LINEMODE_DISCONNECTED:
	g_warning("Disconnected line mode not implemented yet");
	break;
    default:
	g_warning("Invalid line mode %i", mode);
    }
}


static void nxplot_artft_ensure_font(NxplotArtft *self)
{
    if (self->font)
	nxplot_artft_font_unref(self->font);
    self->font = nxplot_artft_get_font(self->current_font_name, self->font_transf_art);
    nxplot_artft_font_get_transf(self->font, &self->font_transf);
}


static void nxplot_artft_invalidate_font(NxplotArtft *self)
{
    if (self->font) {
	nxplot_artft_font_unref(self->font);
	self->font = NULL;
    }
}


static void nxplot_artft_fontname(Nxplot *nxplot, const char *font_name)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);

    g_return_if_fail(font_name);
    self->current_font_name = g_quark_from_string(font_name);
    self->using_symbol_font = (self->current_font_name == symbol_font_name);
    nxplot_artft_invalidate_font(self);
}


static void nxplot_artft_fontsize(Nxplot *nxplot, double size)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(size > 0);
    if (self->state.fontsize != size) {
	self->state.fontsize = size;
	nxplot_artft_update_font_size(self);
    }
}


static void nxplot_artft_textangle(Nxplot *nxplot, double angle)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double new_angle;

    g_return_if_fail(nxplot->opened);
    new_angle = -fmod(angle, 360);
    if (self->state.textangle != new_angle) {
	self->state.textangle = new_angle;
	nxplot_artft_update_font_size(self);
    }
}



/**
 * nxplot_artft_text_line_new:
 * 
 * Create a new empty #NxplotArtftTextLine structure.
 * 
 * Return value: the #NxplotArtftTextLine
 **/
NxplotArtftTextLine* nxplot_artft_text_line_new(void)
{
    NxplotArtftTextLine *tl;
    tl = g_new0(NxplotArtftTextLine, 1);
    tl->mem_chunk = g_mem_chunk_new("NxplotArtftGlyph", sizeof(NxplotArtftGlyph),
				    32*sizeof(NxplotArtftGlyph), G_ALLOC_ONLY);
    tl->glyphs = g_ptr_array_new();
    return tl;
}


/**
 * nxplot_artft_text_line_add_glyph:
 * @tl: the text line
 * 
 * Creates a new glyph and appends it to the text line.
 * 
 * Return value: The newly-allocated and empty glyph.
 **/
__inline__ NxplotArtftGlyph*
nxplot_artft_text_line_add_glyph(NxplotArtftTextLine *tl)
{
    NxplotArtftGlyph *glyph;
    glyph = g_mem_chunk_alloc0(tl->mem_chunk);
    glyph->glyph_bitmap_left = G_MININT;
    g_ptr_array_add(tl->glyphs, glyph);
    return glyph;
}


/**
 * nxplot_artft_text_line_free:
 * @tl: a #NxplotArtftTextLine
 * 
 * Frees a #NxplotArtftTextLine
 **/
void nxplot_artft_text_line_free(NxplotArtftTextLine *tl)
{
    int i;
    NxplotArtftGlyph *glyph;
    for (i = 0; i < tl->glyphs->len; ++i) {
	glyph = g_ptr_array_index(tl->glyphs, i);
	nxplot_artft_cached_glyph_unref(glyph->glyph);
    }
    g_ptr_array_free(tl->glyphs, TRUE);
    g_mem_chunk_destroy(tl->mem_chunk);
    g_free(tl);
}


/**
 * nxplot_artft_text_layout:
 * @self: the plotter
 * @text: the text string
 * 
 * Prepare a line of text for rendering. All glyphs required are
 * loaded, transformed, and positioned along the baseline. The
 * positions and the glyphs are recorded in a structure for later
 * rendering. Some interesting metrics are also recorded. Note: the
 * function stops when a newline is found in the string. So, '\n' is
 * also a string delimiter, besides '\0'.
 * 
 * Return value: a #NxplotArtftTextLine structure
 **/
NxplotArtftTextLine* nxplot_artft_text_layout(NxplotArtft *self, const gchar *text)
{
    int i, glyph_index, last_glyph = 0, error = 0;
    FT_Glyph glyph;
    int  x0 = 0, y0 = 0;
    FT_Vector advance, kerning;
    FT_Face face;
    NxplotArtftTextLine *tl;
    NxplotArtftGlyph *tlglyph;
    NxplotArtftCachedGlyph *cglyph;
    gunichar charcode;
    FT_BBox  cbox;

    g_return_val_if_fail(text, NULL);
    g_return_val_if_fail(g_utf8_validate(text, -1, NULL), NULL);
    last_glyph = 0;

    error++;

    nxplot_artft_ensure_font(self);
    tl = nxplot_artft_text_line_new();
    face = nxplot_artft_font_get_face(self->font);
    tl->face_height = (face->bbox.yMax - face->bbox.yMin) >> 6;

    /* memset(&tl->cbox, 0, sizeof(tl->cbox)); */
    tl->cbox.xMin =  G_MAXINT;
    tl->cbox.yMin =  G_MAXINT;
    tl->cbox.xMax = -G_MAXINT;
    tl->cbox.yMax = -G_MAXINT;

    for (i = 0; (charcode = g_utf8_get_char(text)) && charcode != '\n';
	 i++, text = g_utf8_next_char(text))
    {
	/* replace '-' with minus sign in symbol font */
	if (self->using_symbol_font && charcode == '-')
	    charcode = 0x02212;

	cglyph = nxplot_artft_font_get_glyph(self->font, charcode);
	if (!cglyph) continue;
	tlglyph = nxplot_artft_text_line_add_glyph(tl);
	tlglyph->glyph = cglyph;
	glyph = tlglyph->glyph->glyph;
	glyph_index = tlglyph->glyph->index;
	advance = glyph->advance;
	/* kerning */
	if (last_glyph) {
#ifdef NXPLOT_ARTFT_ENABLE_KERNING
	    error = FT_Get_Kerning(face, last_glyph, glyph_index,
				   0, &kerning);
#else
	    kerning.x = kerning.y = 0;
#endif
	} else
	    kerning.x = kerning.y = 0;
	last_glyph = glyph_index;
	FT_Vector_Transform(&kerning, &self->font_transf);
	kerning.x <<= 10;
	kerning.y <<= 10;
	x0 += kerning.x;
	y0 -= kerning.y;
	tlglyph->pos.x = x0 >> 16;
	tlglyph->pos.y = y0 >> 16;
	x0 += advance.x;
	y0 -= advance.y;

	memset(&cbox, 0, sizeof(cbox));
	FT_Glyph_Get_CBox(glyph, ft_glyph_bbox_subpixels, &cbox);
	tl->cbox.xMin = MIN(tl->cbox.xMin, cbox.xMin);
	tl->cbox.yMin = MIN(tl->cbox.yMin, cbox.yMin);
	tl->cbox.xMax = MAX(tl->cbox.xMax, cbox.xMax);
	tl->cbox.yMax = MAX(tl->cbox.yMax, cbox.yMax);

    } // for

    tl->final_pos.x = x0;
    tl->final_pos.y = y0;

    tl->ascender.x = 0;
    tl->ascender.y = FT_MulFix(face->ascender,
			       face->size->metrics.y_scale);
    FT_Vector_Transform(&tl->ascender, &self->font_transf);

    tl->descender.x = 0;
    tl->descender.y = FT_MulFix(face->descender,
				face->size->metrics.y_scale);
    FT_Vector_Transform(&tl->descender, &self->font_transf);

    return tl;
}




/**
 * nxplot_artft_text_render:
 * @self: the plotter object
 * @tl: the text line
 * @dest_x: destination x coordinate
 * @dest_y: destination y coordinate
 * @foreground: foreground color in libart's RGBA format (alpha is ignored).
 * 
 * Render a text line to screen, starting at the specified
 * coordinates, with the current foreground color. No alignment is
 * performed, so it must have been done previously ( see
 * nxplot_artft_text_align() ).
 **/
void nxplot_artft_text_render(NxplotArtft *self, NxplotArtftTextLine *tl,
			      gint dest_x, gint dest_y, guint32 foreground)
{
    int i, error;
    FT_Glyph        glyph;
    FT_BitmapGlyph  glyph_bitmap;
    FT_Bitmap      *bit;
    guchar *pixels, r, g, b;
    int rowstride, x, y, x0, y0;
    int topspace;
    NxplotArtftGlyph *tlglyph;
    int clip_x1, clip_y1;
    int clip_x2, clip_y2;

    g_return_if_fail(tl);

    nxplot_artft_ensure_font(self);

    pixels = self->pixbuf->pixels;
/*     width  = self->pixbuf->width; */
/*     height = self->pixbuf->height; */
    clip_x1 = self->clip_x;
    clip_x2 = self->clip_x + self->clip_w;
    clip_y1 = self->clip_y;
    clip_y2 = self->clip_y + self->clip_h;
    rowstride = self->pixbuf->rowstride;
    r = (foreground>>24);
    g = ((foreground>>16) & 0xff);
    b = ((foreground>>8) & 0xff);

    

    for (i = 0; i < tl->glyphs->len; ++i)
    {
	/* !< */
	FT_Glyph  glyph_copy;
	/* >! */
	tlglyph = g_ptr_array_index(tl->glyphs, i);
	glyph = tlglyph->glyph->glyph;
	/* !< */
	error = FT_Glyph_Copy(glyph, &glyph_copy);
	glyph = glyph_copy;
	/* >! */
	error = FT_Glyph_To_Bitmap(&glyph, nxplot_artft_render_mode, NULL, TRUE);
	
	if (error) { g_warning("FT_Glyph_To_Bitmap"); continue; }
	// access bitmap content by typecasting
	if (glyph->format != ft_glyph_format_bitmap) {
	    g_warning("glyph->format != ft_glyph_format_bitmap");
	    continue;
	}
	glyph_bitmap = (FT_BitmapGlyph)glyph;
	bit = &glyph_bitmap->bitmap;
	topspace = tl->face_height - glyph_bitmap->top;
	x0 = dest_x + tlglyph->pos.x;
	y0 = dest_y + tlglyph->pos.y;

	if (bit->pixel_mode != ft_pixel_mode_grays) 
	    g_warning("bit->pixel_mode != ft_pixel_mode_grays");
	else {
	    int x0p = x0, y0p = y0 - tl->face_height + topspace;
	    int bitmap_left;

	    if (/*tlglyph->glyph_bitmap_left == G_MININT*/ 1)
		tlglyph->glyph_bitmap_left = glyph_bitmap->left;
	    bitmap_left = tlglyph->glyph_bitmap_left;

	    /* begin draw */
	    for (y = 0; y < bit->rows; y++) 
	    {
		for (x = 0; x < bit->width; x++)
		{
		    register guchar *pixel;
		    register guint16 v, w;
		    int x1 = x0p + x + bitmap_left;
		    int y1 = y0p + y;
		    if (x1 >= clip_x1 && x1 < clip_x2 && y1 >= clip_y1 && y1 < clip_y2) {
			v = bit->buffer[y * bit->pitch + x];
			if (v) {
			    pixel = pixels + rowstride * y1 + x1*3;
			    w = 0xff - v;
			    pixel[0] = (pixel[0]*w + r*v) >> 8;
			    pixel[1] = (pixel[1]*w + g*v) >> 8;
			    pixel[2] = (pixel[2]*w + b*v) >> 8;
			}
		    }
		} /* for x */
	    } /* for y */
	}
	// discard glyph image (bitmap or not)
	FT_Done_Glyph(glyph);
    } // for
}




/**
 * nxplot_artft_text_halign:
 * @tl: the text line
 * @halign: horizontal alignment
 * @x: inout paramter with x coordinate.
 * @y: inout paramter with y coordinate
 * 
 * Adjusts the start coordinates of a text line to achieve a given
 * horizontal alignment.
 * Pre-conditions: (*@x,*@y) has the label coordinates before alignment.
 * Post-conditions: (*@x,*@y) has the adjusted label coordinates after alignment.
 **/
void nxplot_artft_text_halign(NxplotArtftTextLine *tl, NxplotHAlign halign, 
			      double *x, double *y)
{
    ArtPoint pos;

    g_return_if_fail(tl);
    g_return_if_fail(x && y);
    pos.x = *x;
    pos.y = *y;
    /* do horizontal alignment */
    switch (halign)
    {
    case NXPLOT_HALIGN_LEFT:
	break;
    case NXPLOT_HALIGN_CENTER:
	pos.x -= tl->final_pos.x / (2*65536.0);
	pos.y -= tl->final_pos.y / (2*65536.0);
	break;
    case NXPLOT_HALIGN_RIGHT:
	pos.x -= tl->final_pos.x / 65536.0;
	pos.y -= tl->final_pos.y / 65536.0;
	break;
    }
    *x = pos.x;
    *y = pos.y;
}


/**
 * nxplot_artft_text_valign:
 * @tl: the text line
 * @valign: vertical alignment
 * @x: inout parameter with x (window) coordinate
 * @y: inout parameter with y (window) coordinate
 * 
 * Adjusts the start coordinates of a text line to achieve a given
 * vertical alignment.
 * Pre-conditions: (*@x,*@y) has the label coordinates before alignment.
 * Post-conditions: (*@x,*@y) has the adjusted label coordinates after alignment.
 **/
void nxplot_artft_text_valign(NxplotArtftTextLine *tl,
			      NxplotVAlign valign, double *x, double *y)
{
    ArtPoint pos;

    g_return_if_fail(tl);
    g_return_if_fail(x && y);
    pos.x = *x;
    pos.y = *y;
    /* do vertical alignment */
    switch (valign)
    {
    case NXPLOT_VALIGN_BOTTOM:
	pos.x -= tl->descender.x / 64.0;
	pos.y += tl->descender.y / 64.0;
    case NXPLOT_VALIGN_BASELINE:
	break;
    case NXPLOT_VALIGN_CENTER:
	pos.x -= ((tl->ascender.x - tl->descender.x) / 2 + tl->descender.x) / 64.0;
	pos.y += ((tl->ascender.y - tl->descender.y) / 2 + tl->descender.y) / 64.0;
	break;
    case NXPLOT_VALIGN_CAPLINE:	/* Don't know what this is */
    case NXPLOT_VALIGN_TOP:
	pos.x -= tl->ascender.x / 64.0;
	pos.y += tl->ascender.y / 64.0;
	break;
    }
    *x = pos.x;
    *y = pos.y;
}



/**
 * nxplot_artft_update_font_size:
 * @self: -
 * 
 * Set the font size (height) and transformation matrix (angle)
 **/
static void nxplot_artft_update_font_size(NxplotArtft *self)
{
    double size;
    double transf[6];
    double scaling[6];
    double tmp[6];
    size = self->state.fontsize;

    /* Scale the font to the size required in user space coordinates */
    art_affine_scale(scaling, size, size);
    /* Space->window */
    art_affine_multiply(transf, self->space2window, scaling);
    /* Y invert */
    art_affine_scale(tmp, 1.0, -1.0);
    art_affine_multiply(transf, transf, tmp);
    memcpy(self->font_transf_art, transf, 6*sizeof(double));
    nxplot_artft_invalidate_font(self);
}


static void nxplot_artft_label(Nxplot *nxplot, NxplotHAlign halign, 
			       NxplotVAlign valign, const char *text)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    ArtPoint pos;
    NxplotArtftTextLine *tl;
    double last_x = self->state.last_x;
    double last_y = self->state.last_y;
    double x, y, x_novalign, y_novalign;
    double new_x, new_y;

    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);

    if (self->state.textangle != 0) {
	nxplot_savestate(nxplot);
	/* (remember that transformations are in reverse order, always) */
	nxplot_translate(nxplot, last_x, last_y);      /* (3rd) */
	nxplot_rotate(nxplot, -self->state.textangle); /* (2nd) */
	nxplot_translate(nxplot, -last_x, -last_y);    /* (1st) */
    }

    tl = nxplot_artft_text_layout(self, text);

    pos.x = self->state.last_x;
    pos.y = self->state.last_y;
    art_affine_point(&pos, &pos, self->space2window);
    x = pos.x; y = pos.y;
    nxplot_artft_text_halign(tl, halign, &x, &y);
    x_novalign = x;
    y_novalign = y;
    nxplot_artft_text_valign(tl, valign, &x, &y);
    nxplot_artft_text_render(self, tl, x, y, self->state.foreground);

    if (self->state.textangle != 0)
	nxplot_restorestate(nxplot);

    nxplot_artft_window2space(nxplot,
			      x_novalign + tl->final_pos.x / 65536.0,
			      y_novalign + tl->final_pos.y / 65536.0,
			      &new_x, &new_y);
    nxplot_artft_move(nxplot, new_x, new_y);
    nxplot_artft_text_line_free(tl);
}


static NxplotArtftState* nxplot_artft_state_new(void)
{
    return g_new0(NxplotArtftState, 1);
}


static void nxplot_artft_state_free(NxplotArtftState *state)
{
    g_return_if_fail(state);
    g_free(state);
}


static void nxplot_artft_state_copy(NxplotArtftState *dest, NxplotArtftState *src)
{
    g_return_if_fail(dest);
    g_return_if_fail(src);
    memcpy(dest, src, sizeof(NxplotArtftState));
}


static NxplotArtftState * nxplot_artft_state_dup(NxplotArtftState *state)
{
    NxplotArtftState *new_state;
    g_return_val_if_fail(state, NULL);
    new_state = nxplot_artft_state_new();
    nxplot_artft_state_copy(new_state, state);
    return new_state;
}


static void nxplot_artft_savestate(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    /* Note: unlike libplot's savestate/restorestate, artft doesn't
       allow for interrupting a path before savestate and resuming it
       after restorestate. */
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    g_queue_push_tail(self->saved_states, 
		      nxplot_artft_state_dup(&self->state));
}

/**
 * nxplot_artft_state_apply:
 * @nxplot: 
 * 
 * Apply the current state settings, i.e. enforce them, updating other
 * NxplotArtft fields to reflect the current state.
 **/
static void nxplot_artft_state_apply(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    nxplot_artft_update_affine(self);
    nxplot_artft_update_font_size(self);
    update_wlinewidth(self);
    update_dashes(self);
    update_clipping(self);
}


static void nxplot_artft_clear_states(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    NxplotArtftState *state = NULL;
    g_return_if_fail(nxplot->opened);
    if (self->inpath) nxplot_endpath(nxplot);
    while (!g_queue_is_empty(self->saved_states))
	state = g_queue_pop_tail(self->saved_states);
    if (state) {
	nxplot_artft_state_copy(&self->state, state);
	nxplot_artft_state_free(state);
	nxplot_artft_state_apply(nxplot);
    }
}


static void nxplot_artft_restorestate(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    NxplotArtftState *state;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(!g_queue_is_empty(self->saved_states));
    if (self->inpath) nxplot_endpath(nxplot);
    state = g_queue_pop_tail(self->saved_states);
    nxplot_artft_state_copy(&self->state, state);
    nxplot_artft_state_free(state);
    nxplot_artft_state_apply(nxplot);
/*     g_message("NxplotArft(%p): states queue length: %i", */
/* 	      self, self->saved_states->length); */
}


static void nxplot_artft_setmatrix(Nxplot *nxplot, double m0, double m1,
				   double m2, double m3, double tx, double ty)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    self->state.space2unit[0] = m0;
    self->state.space2unit[1] = m1;
    self->state.space2unit[2] = m2;
    self->state.space2unit[3] = m3;
    self->state.space2unit[4] = tx;
    self->state.space2unit[5] = ty;

    update_wlinewidth(self);
    update_dashes(self);
    nxplot_artft_update_affine(self);
    nxplot_artft_update_font_size(self);
}


static void nxplot_artft_concat(Nxplot *nxplot, double m0, double m1,
				double m2, double m3, double tx, double ty)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double tmp[6];
    g_return_if_fail(nxplot->opened);
    tmp[0] = m0;
    tmp[1] = m1;
    tmp[2] = m2;
    tmp[3] = m3;
    tmp[4] = tx;
    tmp[5] = ty;
    art_affine_multiply(self->state.space2unit, self->state.space2unit, tmp);

    update_wlinewidth(self);
    update_dashes(self);
    nxplot_artft_update_affine(self);
    nxplot_artft_update_font_size(self);
}


static void nxplot_artft_rotate(Nxplot *nxplot, double theta)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double tmp[6];

    g_return_if_fail(nxplot->opened);

    art_affine_rotate(tmp, theta);
    art_affine_multiply(self->state.space2unit, tmp, self->state.space2unit);

    update_wlinewidth(self);
    update_dashes(self);
    nxplot_artft_update_affine(self);
    nxplot_artft_update_font_size(self);
}


static void nxplot_artft_scale(Nxplot *nxplot, double sx, double sy)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double tmp[6];

    g_return_if_fail(nxplot->opened);

    art_affine_scale(tmp, sx, sy);
    art_affine_multiply(self->state.space2unit, tmp, self->state.space2unit);

    update_wlinewidth(self);
    update_dashes(self);
    nxplot_artft_update_affine(self);
    nxplot_artft_update_font_size(self);
}


static void nxplot_artft_translate(Nxplot *nxplot, double tx, double ty)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double tmp[6];

    g_return_if_fail(nxplot->opened);

    art_affine_translate(tmp, tx, ty);
    art_affine_multiply(self->state.space2unit, tmp, self->state.space2unit);

/*      update_wlinewidth(self); */
/*      update_dashes(self); */
    nxplot_artft_update_affine(self);
/*      nxplot_artft_update_font_size(self); */
}



static void nxplot_artft_space2window(Nxplot *nxplot, double x, double y,
				      double *wx, double *wy)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    ArtPoint pos;
    g_return_if_fail(wx && wy);
    pos.x = x;
    pos.y = y;
    art_affine_point(&pos, &pos, self->space2window);
    *wx = pos.x;
    *wy = pos.y;
}


static void nxplot_artft_window2space(Nxplot *nxplot, double wx, double wy,
				      double *x, double *y)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    ArtPoint pos;
    g_return_if_fail(x && y);
    pos.x = wx;
    pos.y = wy;
    art_affine_point(&pos, &pos, self->window2space);
    *x = pos.x;
    *y = pos.y;
}



static void nxplot_artft_clear_rect(Nxplot *nxplot, double x1, double y1, double x2, double y2)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    double wx1, wy1, wx2, wy2, t;
    int iwx1, iwy1, iwx2, iwy2, y, span;
    art_u8 r, g, b;

    if (!self->pixbuf) return;

    nxplot_artft_space2window(nxplot, x1, y1, &wx1, &wy1);
    nxplot_artft_space2window(nxplot, x2, y2, &wx2, &wy2);
    if (wx1 > wx2) { t = wx1; wx1 = wx2; wx2 = t; }
    if (wy1 > wy2) { t = wy1; wy1 = wy2; wy2 = t; }
    iwx1 = floor(wx1);
    iwy1 = floor(wy1);
    iwx2 = ceil(wx2);
    iwy2 = ceil(wy2);

    if (iwx1 < 0) iwx1 = 0;
    if (iwx1 >= self->allocation.width) iwx1 = self->allocation.width - 1;

    if (iwx2 < 0) iwx2 = 0;
    if (iwx2 >= self->allocation.width) iwx2 = self->allocation.width - 1;

    if (iwy1 < 0) iwy1 = 0;
    if (iwy1 >= self->allocation.height) iwy1 = self->allocation.height - 1;

    if (iwy2 < 0) iwy2 = 0;
    if (iwy2 >= self->allocation.height) iwy2 = self->allocation.height - 1;

    span = iwx2 - iwx1 + 1;
    if (!span) return;
    r = (self->state.background >> 24);
    g = ((self->state.background>> 16) & 0xff);
    b = ((self->state.background>>  8) & 0xff);

    for (y = iwy1; y <= iwy2; ++y)
	art_rgb_fill_run(self->pixbuf->pixels +
			 self->pixbuf->rowstride*y + 3*iwx1,
			 r, g, b, span);
    if (self->widget)
	gtk_widget_queue_draw_area(self->widget, iwx1, iwy1,
				   iwx2 - iwx1 + 1, iwy2 - iwy1 + 1);
}


/**
 * nxplot_artft_text_get_bbox:
 * @tl: the text line
 * @bbox: the resulting bbox
 * 
 * Compute an aproximated bounding box if the text line would to be
 * rendered at (0,0). To compute the bbox considering a diferent
 * origin (x,y), add (x,y) to (bbox->x, bbox->y).
 **/
void nxplot_artft_text_get_bbox(NxplotArtftTextLine *tl, GdkRectangle *bbox)
{
    ArtPoint  tmp, baseline, asc, desc;
    const ArtPoint pos = {0, 0};
    double x_min, x_max, y_min, y_max;

    g_return_if_fail(tl);
    g_return_if_fail(bbox);
    /* get bounds */
    x_min =  G_MAXDOUBLE;
    x_max = -G_MAXDOUBLE;
    y_min =  G_MAXDOUBLE;
    y_max = -G_MAXDOUBLE;
    /* normalize coordinates */
    baseline.x = tl->final_pos.x / 65536.0;
    baseline.y = tl->final_pos.y / 65536.0;
    asc.x  =  tl->ascender.x / 64.0;
    asc.y  = -tl->ascender.y / 64.0;
    desc.x =  tl->descender.x / 64.0;
    desc.y = -tl->descender.y / 64.0;
#define checkminmax				\
    if (tmp.x < x_min) x_min = tmp.x;		\
    if (tmp.x > x_max) x_max = tmp.x;		\
    if (tmp.y < y_min) y_min = tmp.y;		\
    if (tmp.y > y_max) y_max = tmp.y;

    tmp.x = pos.x;
    tmp.y = pos.y;
    checkminmax;

    tmp.x = pos.x + asc.x;
    tmp.y = pos.y + asc.y;
    checkminmax;

    tmp.x = pos.x + desc.x;
    tmp.y = pos.y + desc.y;
    checkminmax;

    tmp.x = pos.x + baseline.x;
    tmp.y = pos.y + baseline.y;
    checkminmax;

    tmp.x = pos.x + baseline.x + asc.x;
    tmp.y = pos.y + baseline.y + asc.y;
    checkminmax;

    tmp.x = pos.x + baseline.x + desc.x;
    tmp.y = pos.y + baseline.y + desc.y;
    checkminmax;
#undef checkminmax
    bbox->x = floor(x_min);
    bbox->y = floor(y_min);
    bbox->width = ceil(x_max - x_min);
    bbox->height = ceil(y_max - y_min);
}


static void update_clipping(NxplotArtft *self)
{
    double wx1, wy1, wx2, wy2;

    if (self->state.clipping_on) {
/* 	g_message("clipping: x1=%f y1=%f x2=%f y2=%f", */
/* 		  self->state.cliprect.x1, */
/* 		  self->state.cliprect.y1, */
/* 		  self->state.cliprect.x2, */
/* 		  self->state.cliprect.y2); */
	nxplot_artft_space2window(NXPLOT(self),
				  self->state.cliprect.x1,
				  self->state.cliprect.y1,
				  &wx1, &wy1);
	nxplot_artft_space2window(NXPLOT(self),
				  self->state.cliprect.x2,
				  self->state.cliprect.y2,
				  &wx2, &wy2);
	self->clip_w = 1 + (gint) (wx2 - wx1);
	self->clip_h = 1 + (gint) (wy1 - wy2);
	self->clip_x = (gint) wx1;
	self->clip_y = (gint) wy2;
/* 	g_message("clipping: w%i h%i x%i y%i", self->clip_w, self->clip_h, */
/* 		  self->clip_x, self->clip_y); */
	/* clipping buffer is the same is the pixel buffer... */
	*self->clipbuf = *self->pixbuf;
	/* ...except for some 'minor' changes */
	self->clipbuf->width  = self->clip_w;
	self->clipbuf->height = self->clip_h;
	self->clipbuf->pixels += (self->clipbuf->rowstride*self->clip_y +
				  3*self->clip_x);
    } else {
	*self->clipbuf = *self->pixbuf;
	self->clip_w = self->allocation.width;
	self->clip_h = self->allocation.height;
	self->clip_x = 0;
	self->clip_y = 0;
    }
}

static void nxplot_artft_setcliprect(Nxplot *nxplot, NxplotRect *clip_rect)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);

    self->state.cliprect    = *clip_rect;
    self->state.clipping_on = TRUE;
    update_clipping(self);
}

static double nxplot_artft_labelwidth(Nxplot *nxplot, const char *text)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    NxplotArtftTextLine *txtln; 
    double x, y, x0, y0;

    g_return_val_if_fail(nxplot->opened, -1);

    txtln = nxplot_artft_text_layout(self, text);
    g_return_val_if_fail(txtln, 0);
    nxplot_artft_window2space(nxplot,
			      txtln->final_pos.x / 65536.0,
			      txtln->final_pos.y / 65536.0,
			      &x, &y);
    nxplot_artft_window2space(nxplot, 0, 0, &x0, &y0);
    nxplot_artft_text_line_free(txtln);
    x -= x0;
    y -= y0;
#ifdef HAVE_HYPOT
    return hypot(x, y);
#else
    return sqrt(x*x + y*y);
#endif
}


static double nxplot_artft_descender(Nxplot *nxplot)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    FT_Vector descender;
    FT_Face   face;
    double x, y, x0, y0;

    g_return_val_if_fail(nxplot->opened, -1);
    nxplot_artft_ensure_font(self);
    face = nxplot_artft_font_get_face(self->font);
    descender.x = 0;
    descender.y = FT_MulFix(face->descender,
			    face->size->metrics.y_scale);
    FT_Vector_Transform(&descender, &self->font_transf);

    nxplot_artft_window2space(nxplot,
			      descender.x / 64.0,
			      descender.y / 64.0,
			      &x, &y);
    nxplot_artft_window2space(nxplot, 0, 0, &x0, &y0);
    x -= x0;
    y -= y0;
#ifdef HAVE_HYPOT
    return hypot(x, y);
#else
    return sqrt(x*x + y*y);
#endif
}


static void _list_fonts(Nxplot *nxplot, gchar ***fontnames)
{
    *fontnames = nxplot_artft_list_fonts();
}


static void nxplot_artft_labelmetrics(Nxplot             *nxplot,
				      const char         *text,
				      NxplotLabelMetrics *metrics)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    NxplotArtftTextLine *txtln; 
    double x, y, x0, y0;
    FT_Vector descender, ascender;
    FT_Face   face;

    g_return_if_fail(nxplot->opened);
    g_return_if_fail(metrics);

    nxplot_artft_ensure_font(self);
    txtln = nxplot_artft_text_layout(self, text);
    g_return_if_fail(txtln);

    face = nxplot_artft_font_get_face(self->font);
    nxplot_artft_window2space(nxplot, 0, 0, &x0, &y0);

    /* width */
    nxplot_artft_window2space(nxplot,
			      txtln->final_pos.x / 65536.0,
			      txtln->final_pos.y / 65536.0,
			      &x, &y);
    x -= x0;
    y -= y0;
    metrics->width = myhypot(x, y);

    /* descender_max */
    descender.x = 0;
    descender.y = FT_MulFix(face->descender,
			    face->size->metrics.y_scale);
    FT_Vector_Transform(&descender, &self->font_transf);
    nxplot_artft_window2space(nxplot,
			      descender.x / 64.0,
			      descender.y / 64.0,
			      &x, &y);
    x -= x0;
    y -= y0;
    metrics->descender_max = myhypot(x, y);
    
    /* ascender_max */
    ascender.x = 0;
    ascender.y = FT_MulFix(face->ascender,
			   face->size->metrics.y_scale);
    FT_Vector_Transform(&ascender, &self->font_transf);
    nxplot_artft_window2space(nxplot,
			      ascender.x / 64.0,
			      ascender.y / 64.0,
			      &x, &y);
    x -= x0;
    y -= y0;
    metrics->ascender_max = myhypot(x, y);

    /* descender */
    descender.x = 0;
    descender.y = -txtln->cbox.yMin / 64.0;
    nxplot_artft_window2space(nxplot, descender.x, descender.y, &x, &y);
    x -= x0;
    y -= y0;
    metrics->descender = myhypot(x, y);
    if (txtln->cbox.yMin > 0)
	metrics->descender = -metrics->descender;

    /* ascender */
    ascender.x = 0;
    ascender.y = txtln->cbox.yMax / 64.0;
    nxplot_artft_window2space(nxplot, ascender.x, ascender.y, &x, &y);
    x -= x0;
    y -= y0;
    metrics->ascender = myhypot(x, y);
    if (txtln->cbox.yMax < 0)
	metrics->ascender = -metrics->ascender;

    nxplot_artft_text_line_free(txtln);
}


static void nxplot_artft_fillcolor(Nxplot *nxplot, guint16 red, guint16 green, guint16 blue)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    guint32 r, g, b;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    r = red; g = green; b = blue;
    self->state.fillcolor = ((r<<16)&0xff000000) | ((g<<8)&0x00ff0000) | ((b)&0x0000ff00);
}


static void nxplot_artft_fillcolorname(Nxplot *nxplot, const char *name)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    GdkColor color = {0,0,0,0};
    gint rv;
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    
    rv = gdk_color_parse(name, &color);
    if (!rv) g_warning("Parsing of color '%s' failed", name);
    nxplot_artft_fillcolor(nxplot, color.red, color.green, color.blue);
}


static void nxplot_artft_fillopacity(Nxplot *nxplot, guint16 opacity)
{
    NxplotArtft *self = NXPLOT_ARTFT(nxplot);
    g_return_if_fail(nxplot->opened);
    g_return_if_fail(self->pixbuf);
    if (self->inpath)
	nxplot_artft_endpath(nxplot);
    self->state.fillopacity = opacity;
}

/**
 * nxplot_artft_ready:
 * 
 * Check if a NxplotArtft plotter is ready to be used for drawing.
 * 
 **/
gboolean nxplot_artft_ready(NxplotArtft *self)
{
    g_return_val_if_fail(IS_NXPLOT_ARTFT(self), FALSE);
    return self->pixbuf != NULL;
}


GdkPixbuf* nxplot_artft_get_pixbuf(NxplotArtft *self)
{
    GdkPixbuf *pixbuf;
    guchar    *pixels;
    gsize      sz;

    sz = self->pixbuf->rowstride*self->pixbuf->height;
    pixels  = g_new(guchar, sz);
    memcpy(pixels, self->pixbuf->pixels, sz);
    pixbuf = gdk_pixbuf_new_from_data(pixels,
				      GDK_COLORSPACE_RGB,
				      self->pixbuf->has_alpha,
				      self->pixbuf->bits_per_sample,
				      self->pixbuf->width,
				      self->pixbuf->height,
				      self->pixbuf->rowstride,
				      NULL, NULL);
    return pixbuf;
}

